#ifndef OSRM_EXTRACTOR_NAME_TABLE_HPP
#define OSRM_EXTRACTOR_NAME_TABLE_HPP

// Deprecated: use extractor/string_table.hpp instead
#include "extractor/string_table.hpp"

#endif // OSRM_EXTRACTOR_NAME_TABLE_HPP
