include "waypoint.fbs";
namespace osrm.engine.api.fbresult;

table TableResult {
    durations: [float];
    rows: ushort;
    cols: ushort;
    distances: [float];
    destinations: [Waypoint];
    fallback_speed_cells: [uint];
}